aboutsummaryrefslogtreecommitdiff
path: root/src/DSi_NAND.cpp
diff options
context:
space:
mode:
Diffstat (limited to 'src/DSi_NAND.cpp')
-rw-r--r--src/DSi_NAND.cpp112
1 files changed, 88 insertions, 24 deletions
diff --git a/src/DSi_NAND.cpp b/src/DSi_NAND.cpp
index 4366ce6..912fee4 100644
--- a/src/DSi_NAND.cpp
+++ b/src/DSi_NAND.cpp
@@ -1,5 +1,5 @@
/*
- Copyright 2016-2021 Arisotura
+ Copyright 2016-2022 melonDS team
This file is part of melonDS.
@@ -49,8 +49,48 @@ UINT FF_ReadNAND(BYTE* buf, LBA_t sector, UINT num);
UINT FF_WriteNAND(BYTE* buf, LBA_t sector, UINT num);
-bool Init(FILE* nandfile, u8* es_keyY)
+bool Init(u8* es_keyY)
{
+ CurFile = nullptr;
+
+ std::string nandpath = Platform::GetConfigString(Platform::DSi_NANDPath);
+ std::string instnand = nandpath + Platform::InstanceFileSuffix();
+
+ FILE* nandfile = Platform::OpenLocalFile(instnand, "r+b");
+ if ((!nandfile) && (Platform::InstanceID() > 0))
+ {
+ FILE* orig = Platform::OpenLocalFile(nandpath, "rb");
+ if (!orig)
+ {
+ printf("Failed to open DSi NAND\n");
+ return false;
+ }
+
+ fseek(orig, 0, SEEK_END);
+ long len = ftell(orig);
+ fseek(orig, 0, SEEK_SET);
+
+ nandfile = Platform::OpenLocalFile(instnand, "w+b");
+ if (nandfile)
+ {
+ u8* tmpbuf = new u8[0x10000];
+ for (long i = 0; i < len; i+=0x10000)
+ {
+ long blklen = 0x10000;
+ if ((i+blklen) > len) blklen = len-i;
+
+ fread(tmpbuf, blklen, 1, orig);
+ fwrite(tmpbuf, blklen, 1, nandfile);
+ }
+ delete[] tmpbuf;
+ }
+
+ fclose(orig);
+ fclose(nandfile);
+
+ nandfile = Platform::OpenLocalFile(instnand, "r+b");
+ }
+
if (!nandfile)
return false;
@@ -138,10 +178,17 @@ void DeInit()
f_unmount("0:");
ff_disk_close();
+ if (CurFile) fclose(CurFile);
CurFile = nullptr;
}
+FILE* GetFile()
+{
+ return CurFile;
+}
+
+
void GetIDs(u8* emmc_cid, u64& consoleid)
{
memcpy(emmc_cid, eMMC_CID, 16);
@@ -919,29 +966,44 @@ bool CreateSaveFile(const char* path, u32 len)
u32 clustersize, maxfiles, totsec16, fatsz16;
// CHECKME!
- // code inspired from https://github.com/JeffRuLz/TMFH/blob/master/arm9/src/sav.c
- if (len < 573440)
- {
- clustersize = 512;
- maxfiles = 16;
- }
- else if (len < 5472256)
+ // code inspired from https://github.com/Epicpkmn11/NTM/blob/master/arm9/src/sav.c
+ const u16 sectorsize = 0x200;
+
+ // fit maximum sectors for the size
+ const u16 maxsectors = len / sectorsize;
+ u16 tracksize = 1;
+ u16 headcount = 1;
+ u16 totsec16next = 0;
+ while (totsec16next <= maxsectors)
{
- clustersize = 2048;
- maxfiles = 256;
+ totsec16next = tracksize * (headcount + 1) * (headcount + 1);
+ if (totsec16next <= maxsectors)
+ {
+ headcount++;
+ totsec16 = totsec16next;
+
+ tracksize++;
+ totsec16next = tracksize * headcount * headcount;
+ if (totsec16next <= maxsectors)
+ {
+ totsec16 = totsec16next;
+ }
+ }
}
- else
+ totsec16next = (tracksize + 1) * headcount * headcount;
+ if (totsec16next <= maxsectors)
{
- clustersize = 4096;
- maxfiles = 256;
+ tracksize++;
+ totsec16 = totsec16next;
}
- if (len <= 0x4000) fatsz16 = 1;
- else if (len <= 0x200000) fatsz16 = 3;
- else fatsz16 = 6;
+ maxfiles = len < 0x8C000 ? 0x20 : 0x200;
+ clustersize = (totsec16 > (8 << 10)) ? 8 : (totsec16 > (1 << 10) ? 4 : 1);
- if (len == 0x4000) totsec16 = 27;
- else totsec16 = len >> 9;
+ #define ALIGN(v, a) (((v) % (a)) ? ((v) + (a) - ((v) % (a))) : (v))
+ u16 totalclusters = ALIGN(totsec16, clustersize) / clustersize;
+ u32 fatbytes = (ALIGN(totalclusters, 2) / 2) * 3; // 2 sectors -> 3 byte
+ fatsz16 = ALIGN(fatbytes, sectorsize) / sectorsize;
FF_FIL file;
FRESULT res;
@@ -960,17 +1022,19 @@ bool CreateSaveFile(const char* path, u32 len)
// create FAT header
data[0x000] = 0xE9;
memcpy(&data[0x003], "MSWIN4.1", 8);
- *(u16*)&data[0x00B] = 512; // bytes per sector
- data[0x00D] = clustersize >> 9;
+ *(u16*)&data[0x00B] = sectorsize; // bytes per sector
+ data[0x00D] = clustersize;
*(u16*)&data[0x00E] = 1; // reserved sectors
data[0x010] = 2; // num FATs
- *(u16*)&data[0x011] = maxfiles << 1;
+ *(u16*)&data[0x011] = maxfiles;
*(u16*)&data[0x013] = totsec16;
data[0x015] = 0xF8;
*(u16*)&data[0x016] = fatsz16;
- data[0x024] = 0x07;
+ *(u16*)&data[0x018] = tracksize;
+ *(u16*)&data[0x01A] = headcount;
+ data[0x024] = 0x05;
data[0x026] = 0x29;
- *(u32*)&data[0x027] = 305419896;
+ *(u32*)&data[0x027] = 0x12345678;
memcpy(&data[0x02B], "VOLUMELABEL", 11);
memcpy(&data[0x036], "FAT12 ", 8);
*(u16*)&data[0x1FE] = 0xAA55;